/**CFile***********************************************************************

  FileName    [statenode.c]

  PackageName [mc]

  Synopsis    [The main routines of the <tt>statenode</tt> data structure.]

  Description [The <tt>statenode</tt> implements a data structure
  to use to manage tree-like annotated counter-examples.]

  Author      [Simon Busard]

  Copyright   [
  This file is part of the ``mc'' package of NuSMV version 2.
  Copyright (C) 1998-2001 by CMU and ITC-irst.

  NuSMV version 2 is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.

  For more information of NuSMV see <http://nusmv.irst.itc.it>
  or email to <nusmv-users@irst.itc.it>.
  Please report bugs to <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to <nusmv@irst.itc.it>. ]

******************************************************************************/

#include "statenode.h"

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

int id_node = 0;

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static void print_dtd ARGS((FILE *));
static void print_statenode ARGS((BddFsm_ptr,BddEnc_ptr,statenode_ptr,node_ptr,FILE *));
static void print_branch ARGS((BddFsm_ptr,BddEnc_ptr,node_ptr, FILE *));
static void print_combinatorial ARGS((BddFsm_ptr,BddEnc_ptr,bdd_ptr,bdd_ptr, FILE *));
static void print_input ARGS((BddFsm_ptr,BddEnc_ptr,bdd_ptr,bdd_ptr, FILE *));

static void remove_unitary_branches ARGS((statenode_ptr));

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis           [Creates a new state node.]

  Description        [This function creates a state node containing a state,
  an annotation, an id and a list of branches.  The returned data structure
  is a LISP-like tree.]

  SideEffects        []

******************************************************************************/
statenode_ptr statenode_new(bdd_ptr state)
{
	statenode_ptr statenode = cons((node_ptr) state,
									cons(Nil,
											cons(Nil,
													cons(Nil,Nil))));
	statenode_set_id(statenode, 0);
	return statenode;
}

/**Function********************************************************************

  Synopsis           [Duplicates a given state.]

  Description        [This function duplicates the given state.  It only
  duplicates the state.  The rest of the returned node is reinitialized.]

  SideEffects        []

******************************************************************************/
statenode_ptr statenode_dup(statenode_ptr self)
{
	statenode_ptr statenode = statenode_new(bdd_dup((bdd_ptr)statenode_get_state(self)));
	return statenode;
}

/**Function********************************************************************

  Synopsis           [Free a given state.]

  Description        [This function frees the memory allocated to a given state.
  It uses the functionalities given by the node package.]

  SideEffects        []

******************************************************************************/
void statenode_free(statenode_ptr self)
{
	// Free loop
	if(cdr(cdr(cdr(cdr(self)))) != Nil)
		free_node(cdr(cdr(cdr(cdr(self)))));
	// Free branches
	walk(statenode_free_branch, car(cdr(cdr(cdr(self)))));
	free_node(cdr(cdr(cdr(self))));
	// Free annotation
	// free_node(car(cdr(cdr(self)))); // No need to free the spec!
	free_node(cdr(cdr(self)));
	// Free id
	free_node(cdr(self));
	// Free state and statenode
	bdd_free(dd_manager, (bdd_ptr) car(self));
	free_node(self);
}

/**Function********************************************************************

  Synopsis           [Free a given branch.]

  Description        [This function frees the memory allocated to a given branch.
  It uses the functionalities given by the node package.  A branch is a list
  of state nodes.]

  SideEffects        []

******************************************************************************/
void statenode_free_branch (node_ptr branch)
{
	node_ptr tmp;

	if(branch == Nil)
		return;

	// Free first node
	statenode_free(car(branch));
	tmp = branch;
	branch = cdr(branch);
	free_node(tmp);

	while(branch != Nil) {
		// Free input
		bdd_free(dd_manager, (bdd_ptr) car(branch));
		tmp = branch;
		branch = cdr(branch);
		free_node(tmp);

		// Free node
		statenode_free(car(branch));
		tmp = branch;
		branch = cdr(branch);
		free_node(tmp);
	}
}

/**Function********************************************************************

  Synopsis           [Removes a branch from a statenode's branches.]

  Description        [This function removes a branch from node's list of branches
  if the branch is contained in this list, it does nothing otherwise.]

  SideEffects        []

******************************************************************************/
void statenode_remove_branch (statenode_ptr node, node_ptr branch)
{
	node_ptr branches = statenode_get_branches(node);
	node_ptr prev = Nil;

	while(branches != Nil)
	{
		if(car(branches) == branch)
		{
			if(prev == Nil)
			{
				statenode_set_branches(node, cdr(branches));
			}
			else
			{
				node_node_setcdr(prev,cdr(branches));
			}
			return;
		}
		prev = branches;
		branches = cdr(branches);
	}
}

/**Function********************************************************************

  Synopsis           [Returns the state loop.]

  Description        [If it exists, returns the state node to loop to from
  the given state node.  If such a state node doesn't exist, returns Nil.]

  SideEffects        []

******************************************************************************/
statenode_ptr statenode_get_loop(statenode_ptr self)
{
	if(cdr(cdr(cdr(cdr(self)))) != Nil)
		return car(cdr(cdr(cdr(cdr(self)))));
	else
		return Nil;
}

/**Function********************************************************************

  Synopsis           [Sets the state loop.]

  Description        [Sets the state node to loop to from self to loop_to.]

  SideEffects        []

******************************************************************************/
void statenode_set_loop(statenode_ptr self, statenode_ptr loop_to)
{
	if(cdr(cdr(cdr(cdr(self)))) != Nil)
		node_node_setcar(cdr(cdr(cdr(cdr(self)))), loop_to);
	else
		node_node_setcdr(cdr(cdr(cdr(self))), cons(loop_to, Nil));
}

/**Function********************************************************************

  Synopsis           [Returns whether the given state node satisfies the given
  specification.]

  Description        [Returns whether the given state node satisfies the given
  specification.  The specification is defined by specification and context.]

  SideEffects        []

******************************************************************************/
boolean statenode_satisfies(statenode_ptr self, BddFsm_ptr fsm, BddEnc_ptr enc,
							node_ptr specification, node_ptr context)
{
	bdd_ptr s;
	boolean res = false;

	s = eval_spec(fsm, enc, specification, context);
	bdd_and_accumulate(dd_manager, &s, statenode_get_state(self));

	if(bdd_isnot_zero(dd_manager, s))
		res = true;

	bdd_free(dd_manager, s);

	return res;
}

/**Function********************************************************************

  Synopsis           [Displays a tree-like annotated counter-example starting
  at path.]

  Description        [Displays a tree-like annotated counter-example starting
  at path.  The output is formated with XML.]

  SideEffects        []

******************************************************************************/
void statenode_display(BddFsm_ptr fsm, BddEnc_ptr enc, statenode_ptr path, node_ptr spec, FILE * output)
{
	// Remove orphan branches
	remove_unitary_branches(path);

	reset_indent_size();

	indent(output);
	fprintf(output, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
	//print_dtd(output);
	indent(output);
	fprintf(output, "<counterexample specification=\"");
	print_node_XML(output, spec);
	fprintf(output, "\">\n");

	inc_indent_size();

	id_node = 0;

	print_statenode(fsm, enc, car(path), Nil, output);
	print_branch(fsm, enc, path, output);

	dec_indent_size();

	indent(output);
	fprintf(output, "</counterexample>\n");
}

/**Function********************************************************************

  Synopsis           [Returns the total number of states contained in the
  given path.]

  Description        [Returns the total number of states contained in the
  given path.  This function recursively adds states from sub-branches.
  Notice that all statenodes are counted.  So are first ones of sub-branches.]

  SideEffects        []

******************************************************************************/
int statenode_get_state_count(node_ptr path)
{
	int count = 0;
	node_ptr branches;

	while(path != Nil)
	{
		count++;
		branches = statenode_get_branches(car(path));

		while(branches != Nil)
		{
			count += statenode_get_state_count(car(branches));

			branches = cdr(branches);
		}

		path = cdr(path);
		if(path != Nil)
			path = cdr(path);
	}
	return count;
}

/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis           [Prints the DTD describing a valid counterexample XML
  file.]

  Description        [Prints the DTD describing a valid counterexample XML
  file.]

  SideEffects        []

******************************************************************************/
static void print_dtd(FILE * output)
{
	indent(output);
	fprintf(output,"<!DOCTYPE counterexample [\n");
	indent(output);
	fprintf(output,"<!ELEMENT counterexample (node)+>\n");
	indent(output);
	fprintf(output,"<!ATTLIST counterexample specification CDATA #REQUIRED>\n");
	indent(output);
	fprintf(output,"<!ELEMENT node (combinatorial?,input?,(state|loop),annotation,branch*)>\n");
	indent(output);
	fprintf(output,"<!ELEMENT combinatorial (value)*>\n");
	indent(output);
	fprintf(output,"<!ELEMENT input (value)*>\n");
	indent(output);
	fprintf(output,"<!ELEMENT state (value)*>\n");
	indent(output);
	fprintf(output,"<!ATTLIST state id CDATA #REQUIRED>\n");
	indent(output);
	fprintf(output,"<!ELEMENT loop EMPTY>\n");
	indent(output);
	fprintf(output,"<!ATTLIST loop id CDATA #REQUIRED>\n");
	indent(output);
	fprintf(output,"<!ELEMENT annotation (#PCDATA)>\n");
	indent(output);
	fprintf(output,"<!ELEMENT branch (node)+>\n");
	indent(output);
	fprintf(output,"<!ATTLIST branch specification CDATA #IMPLIED>\n");
	indent(output);
	fprintf(output,"<!ELEMENT value (#PCDATA)>\n");
	indent(output);
	fprintf(output,"<!ATTLIST value variable CDATA #REQUIRED>\n");
	indent(output);
	fprintf(output,"]>\n");
}

/**Function********************************************************************

  Synopsis           [Prints a particular state node.]

  Description        [Prints a particular state node.  It displays its fields
  but also all the branches starting at the state node.]

  SideEffects        []

******************************************************************************/
static void print_statenode(BddFsm_ptr fsm, BddEnc_ptr enc, statenode_ptr statenode, node_ptr input, FILE * output)
{
	node_ptr branches;
	ListIter_ptr los;
	NodeList_ptr list;
	add_ptr add_state;

	Encoding_ptr senc = BddEnc_get_symbolic_encoding(enc);
	DdManager* dd = BddEnc_get_dd_manager(enc);

	indent(output);
	fprintf(output, "<node>\n");
	inc_indent_size();

	if(statenode_get_id(statenode) == 0)
	{
		statenode_set_id(statenode, ++id_node);
	}

	// Input and combinatorial
	if(input != Nil)
	{
		print_combinatorial(fsm, enc, statenode_get_state(statenode), (bdd_ptr) input, output);
		print_input(fsm, enc, statenode_get_state(statenode), (bdd_ptr) input, output);
	}

	if(statenode_get_loop(statenode) != Nil)
	{
		// Print loop
		indent(output);
		fprintf(output, "<loop to=\"%d\" />\n", statenode_get_id(statenode_get_loop(statenode)));
	}
	else
	{
		// Print state
		indent(output);
		fprintf(output, "<state id=\"%d\">\n", statenode_get_id(statenode));
		inc_indent_size();

		// Print values
		add_state = bdd_to_add(dd, (BddStates) statenode_get_state(statenode));
		list = Encoding_get_model_state_symbols_list(senc);
		los = NodeList_get_first_iter(list);
		while (! ListIter_is_end(los))
		{
			node_ptr cur_sym_value;
			node_ptr cur_sym = NodeList_get_value_at(list, los);
			add_ptr cur_sym_vals = BddEnc_expr_to_add(enc, cur_sym, Nil);
			add_ptr tmp = add_if_then(dd, add_state, cur_sym_vals);

			cur_sym_value = add_value(dd, tmp);

			add_free(dd, tmp);
			add_free(dd, cur_sym_vals);

			indent(output);
			fprintf(output,
					"<value variable=\"%s\">%s</value>\n",
					sprint_node_XML(cur_sym),
					sprint_node_XML(cur_sym_value));

			los = ListIter_get_next(los);
		}

		dec_indent_size();
		indent(output);
		fprintf(output, "</state>\n");

		//Print annotation
		indent(output);
		fprintf(output, "<annotation>");
		if(statenode_get_annotation(statenode) != Nil) {
			print_node_XML(output, statenode_get_annotation(statenode));
		}
		fprintf(output, "</annotation>\n");

		// Print branches
		branches = statenode_get_branches(statenode);
		while(branches != Nil)
		{
			if(llength(car(branches)) > 1)
			{
				statenode_set_id(car(car(branches)), statenode_get_id(statenode));

				indent(output);
				fprintf(output, "<branch specification=\"");
				print_node_XML(output, statenode_get_annotation(car(car(branches))));
				fprintf(output, "\">\n");
				inc_indent_size();

				// Print branch
				print_branch(fsm, enc, car(branches), output);

				dec_indent_size();
				indent(output);
				fprintf(output, "</branch>\n");

			}
			branches = cdr(branches);
		}
	}

	dec_indent_size();
	indent(output);
	fprintf(output, "</node>\n");
}

/**Function********************************************************************

  Synopsis           [Prints a particular branch.]

  Description        [Prints a particular branch.  It displays every state node
  contained in the given branch.  It discards the first state node.]

  SideEffects        []

******************************************************************************/
static void print_branch(BddFsm_ptr fsm, BddEnc_ptr enc, node_ptr branch, FILE * output)
{
	node_ptr input;
	statenode_ptr statenode;

	branch = cdr(branch);

	while(branch != Nil)
	{
		input = car(branch);
		branch = cdr(branch);
		statenode = (statenode_ptr) car(branch);
		branch = cdr(branch);
		print_statenode(fsm, enc, statenode, input, output);
	}
}

/**Function********************************************************************

  Synopsis           [Prints combinatorial state inputs.]

  Description        [Prints combinatorial state inputs.]

  SideEffects        []

******************************************************************************/
static void print_combinatorial(BddFsm_ptr fsm, BddEnc_ptr enc, bdd_ptr state, bdd_ptr input, FILE * output)
{
	ListIter_ptr los;
	NodeList_ptr list;
	BddInputs curr_input = (BddInputs) input;

	bdd_ptr bdd_comb;
	add_ptr add_input, add_comb;

	Encoding_ptr senc = BddEnc_get_symbolic_encoding(enc);
	DdManager* dd = BddEnc_get_dd_manager(enc);

	bdd_comb = bdd_and(dd, state, curr_input);
	add_comb = bdd_to_add(dd, bdd_comb);
	bdd_free(dd, bdd_comb);

	indent(output);
	fprintf(output,"<combinatorial>\n");
	inc_indent_size();

	list = Encoding_get_model_state_input_symbols_list(senc);
	los = NodeList_get_first_iter(list);
	while (! ListIter_is_end(los)) {
		node_ptr cur_sym_value;
		node_ptr cur_sym = NodeList_get_value_at(list, los);
		add_ptr cur_sym_vals = BddEnc_expr_to_add(enc, cur_sym, Nil);
		add_ptr tmp = add_if_then(dd, add_comb, cur_sym_vals);

		cur_sym_value = add_value(dd, tmp);

		add_free(dd, tmp);
		add_free(dd, cur_sym_vals);

		indent(output);
		fprintf(output,
			"<value variable=\"%s\">%s</value>\n",
			sprint_node_XML(cur_sym),
			sprint_node_XML(cur_sym_value));

		los = ListIter_get_next(los);
	}
	add_free(dd, add_comb);

	dec_indent_size();
	indent(output);
	fprintf(output,"</combinatorial>\n");
}

/**Function********************************************************************

  Synopsis           [Prints state inputs.]

  Description        [Prints states inputs.]

  SideEffects        []

******************************************************************************/
static void print_input(BddFsm_ptr fsm, BddEnc_ptr enc, bdd_ptr state, bdd_ptr input, FILE * output)
{

	ListIter_ptr los;
	NodeList_ptr list;
	BddInputs curr_input = (BddInputs) input;

	bdd_ptr bdd_comb;
	add_ptr add_input, add_comb;

	Encoding_ptr senc = BddEnc_get_symbolic_encoding(enc);
	DdManager* dd = BddEnc_get_dd_manager(enc);

	bdd_comb = bdd_and(dd, state, curr_input);
	add_comb = bdd_to_add(dd, bdd_comb);
	bdd_free(dd, bdd_comb);

	indent(output);
	fprintf(output,"<input>\n");
	inc_indent_size();

	add_input = bdd_to_add(dd, curr_input);
	list = Encoding_get_model_input_symbols_list(senc);
	los = NodeList_get_first_iter(list);
	while (! ListIter_is_end(los)) {
		node_ptr cur_sym_value;
		node_ptr cur_sym = NodeList_get_value_at(list, los);
		add_ptr cur_sym_vals = BddEnc_expr_to_add(enc, cur_sym, Nil);
		add_ptr tmp = add_if_then(dd, add_input, cur_sym_vals);

		cur_sym_value = add_value(dd, tmp);

		add_free(dd, tmp);
		add_free(dd, cur_sym_vals);

		indent(output);
		fprintf(output,
				"<value variable=\"%s\">%s</value>\n",
				sprint_node_XML(cur_sym),
				sprint_node_XML(cur_sym_value));

		los = ListIter_get_next(los);
	}

	dec_indent_size();
	indent(output);
	fprintf(output,"</input>\n");
}

/**Function********************************************************************

  Synopsis           [Checks every state branch and removes the ones that are
  composed of only one state.]

  Description        [Checks every state branch and removes the ones that are
  composed of only one state.  If a state has a branch with only one state,
  each branch of the only state of this branch are moved to its parent.]

  SideEffects        []

******************************************************************************/
static void remove_unitary_branches(statenode_ptr path)
{
	statenode_ptr s, subs;
	node_ptr branches, b, sub_branches, sb;

	while(path != Nil)
	{
		s = car(path);
		branches = statenode_get_branches(s);

		while(branches != Nil)
		{
			b = car(branches);

			remove_unitary_branches(b);

			subs = car(b);
			sub_branches = statenode_get_branches(subs);

			while(sub_branches != Nil)
			{
				sb = car(sub_branches);
				statenode_remove_branch(subs, sb);
				statenode_add_branch(s, sb);
				sub_branches = cdr(sub_branches);
			}

			if(llength(b) == 1)
			{
				statenode_remove_branch(s, b);
				statenode_free_branch(b);
			}

			branches = cdr(branches);
		}

		path = cdr(path);
		if(path != Nil)
			path = cdr(path);
	}
}
