package counterexample;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
 * Branch is the internal representation of a branch in a counterexample.
 * 
 * A branch is a list of states linked by transitions, explaining why a given
 * state (the source of the branch) satisfies a given formula.
 * 
 * @author Simon Busard <simon.busard@uclouvain.be>
 * 
 */
public class Branch {

	/**
	 * The formula that the branch satisfies.
	 */
	private String annotation;

	/**
	 * The first transition of the branch.
	 */
	private Transition initial;

	/**
	 * Creates a new branch annotated with annotation and starting with initial.
	 * 
	 * @param annotation
	 *            the annotation of the branch.
	 * @param initial
	 *            the first transition of the branch.
	 */
	public Branch(String annotation, Transition initial) {
		this.annotation = annotation;
		this.initial = initial;
	}

	/**
	 * @return the annotation of the branch.
	 */
	public String getAnnotation() {
		return annotation;
	}

	/**
	 * @return the first transition of the branch.
	 */
	public Transition getInitial() {
		return initial;
	}

	/**
	 * Returns the set of states belonging to this branch. The states of
	 * sub-branches are included.
	 * 
	 * @return the set of states belonging to this branch.
	 */
	public Set<State> getAllStates() {
		Set<State> states = new HashSet<State>();
		Transition tr = this.getInitial();
		while (tr != null && !states.contains(tr.getTo())
				&& !tr.getTo().equals(this.getInitial().getFrom())) {
			states.add(tr.getTo());
			for (Branch branch : tr.getTo().getBranches()) {
				states.addAll(branch.getAllStates());
			}
			tr = tr.getTo().getNext();
		}
		return states;
	}

	/**
	 * Returns the list of states belonging to this branch. The states of
	 * sub-branches are not included.
	 * 
	 * @return the list of states belonging to this branch.
	 */
	public List<State> getStates() {
		List<State> states = new ArrayList<State>();
		Transition tr = this.getInitial();
		while (tr != null && !states.contains(tr.getTo())
				&& !tr.getTo().equals(this.getInitial().getFrom())) {
			states.add(tr.getTo());
			tr = tr.getTo().getNext();
		}
		return states;
	}
}
