package counterexample;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.xml.sax.SAXException;

import xml.XMLLoader;

/**
 * CounterexampleManager is a central counterexamples store.
 * 
 * @author Simon Busard <simon.busard@uclouvain.be>
 * 
 */

public class CounterexampleManager {

	/**
	 * The list of counterexamples this manager stores.
	 */
	private List<Counterexample> counterexamples;

	/**
	 * The XML loader used to get a counterexample stored in an XML format.
	 */
	private XMLLoader loader;

	/**
	 * Creates a new counterexamples manager using loader to get an internal
	 * representation of a counterexample stored in an XML format.
	 * 
	 * @param loader
	 *            the engine to get a counterexample stored in an XML file.
	 */
	public CounterexampleManager(XMLLoader loader) {
		this.counterexamples = new ArrayList<Counterexample>();
		this.loader = loader;
	}

	/**
	 * Stores the counterexample stored in the given XML file. Returns the id of
	 * the stored counterexample.
	 * 
	 * @param XMLfilepath
	 *            the path to a file containing an XML representation of a
	 *            counterexample.
	 * 
	 * @return the id of the stored counterexample.
	 * @throws IOException
	 *             an IOException is thrown if an error occurred while reading
	 *             the given file.
	 * @throws SAXException
	 *             an SAXException is thrown if an error occurred while parsing
	 *             the given file.
	 */
	public int storeCounterexample(String XMLfilepath) throws SAXException,
			IOException {
		return this.storeCounterexample(loader.getCounterexample(XMLfilepath));
	}

	/**
	 * Stores the given counterexample. Returns the id of the stored
	 * counterexample.
	 * 
	 * @param counterexample
	 *            the counterexample to store.
	 * @return the id of the stored counterexample.
	 */
	public int storeCounterexample(Counterexample counterexample) {
		this.counterexamples.add(counterexample);
		return this.counterexamples.size() - 1;
	}

	/**
	 * Stores the counterexample stored in a given file.
	 * 
	 * @param selectedFile
	 *            the file containing the XML representation of a
	 *            counterexample.
	 * @return the id of the stored counterexample.
	 * @throws IOException
	 *             an IOException is thrown if an error occurred while reading
	 *             the given file.
	 * @throws SAXException
	 *             an SAXException is thrown if an error occurred while parsing
	 *             the given file.
	 */
	public int storeCounterexample(File selectedFile) throws SAXException,
			IOException {
		return this.storeCounterexample(loader.getCounterexample(selectedFile));
	}

	/**
	 * Returns the stored counterexample with counterexampleId as id if it
	 * exists, null otherwise.
	 * 
	 * @param counterexampleId
	 *            the id of the requested counterexample.
	 * @return the stored counterexample with counterexampleId as id if it
	 *         exists, null otherwise.
	 */
	public Counterexample getCounterexample(int counterexampleId) {
		return (0 <= counterexampleId && counterexampleId < this.counterexamples
				.size()) ? this.counterexamples.get(counterexampleId) : null;
	}

}
