package counterexample;

import java.util.ArrayList;
import java.util.List;

/**
 * State is the internal representation of a state.
 * 
 * A state has an id and a set of variable-value pairs, is annotated with a
 * formula that it satisfies and possesses a (possibly null) outgoing transition
 * and a (possibly empty) set of branches.
 * 
 * @author Simon Busard <simon.busard@uclouvain.be>
 * 
 */

public class State {

	/**
	 * The id of this state.
	 */
	private int id;

	/**
	 * The list of variable-value pairs of this state.
	 */
	private Value[] values;

	/**
	 * The formula that this state satisfies.
	 */
	private String annotation;

	/**
	 * The outgoing transition.
	 */
	private Transition next;

	/**
	 * The ingoing transition.
	 */
	private Transition prev;

	/**
	 * The list of branches starting at this state.
	 */
	private List<Branch> branches;

	/**
	 * Creates a new state with id as id, values as set of variable-value pairs
	 * of the created state and annotation as annotation. Its outgoing
	 * transition and list of branches are empty.
	 * 
	 * @param id
	 *            the id of the created state.
	 * @param values
	 *            a list of variable-value pairs.
	 * @param annotation
	 *            the formula the state satisfies.
	 */
	public State(int id, Value[] values, String annotation) {
		this.id = id;
		this.values = values;
		this.annotation = annotation;
		this.next = null;
		this.prev = null;
		this.branches = new ArrayList<Branch>();
	}

	/**
	 * @return the id.
	 */
	public int getId() {
		return id;
	}

	/**
	 * @return the variable-value pairs.
	 */
	public Value[] getValues() {
		return values;
	}

	/**
	 * @return the annotation.
	 */
	public String getAnnotation() {
		return annotation;
	}

	/**
	 * @return the set of branches starting at this state.
	 */
	public Branch[] getBranches() {
		return branches.toArray(new Branch[branches.size()]);
	}

	/**
	 * @return the outgoing transitions.
	 */
	public Transition getNext() {
		return next;
	}

	/**
	 * @return the ingoing transition (ignoring loops).
	 */
	public Transition getPrev() {
		return prev;
	}

	/**
	 * Adds a branch to the set of branches of this state.
	 * 
	 * @param branch
	 *            the branch to add.
	 */
	public void addBranch(Branch branch) {
		this.branches.add(branch);
	}

	/**
	 * Sets the outgoing transition to next.
	 * 
	 * @param next
	 *            the new outgoing transition.
	 */
	public void setNext(Transition next) {
		this.next = next;
	}

	/**
	 * Sets the ingoing transition to prev.
	 * 
	 * @param prev
	 *            the new ingoing transition.
	 */
	public void setPrev(Transition prev) {
		this.prev = prev;
	}

	/**
	 * Returns the set of outgoing transitions.
	 * 
	 * @return The set of outgoing transitions. If this state has no outgoing
	 *         transition, the returned array is empty (length = 0).
	 */
	public Transition[] getNexts() {

		ArrayList<Transition> transitions = new ArrayList<Transition>();

		if (this.next != null) {
			transitions.add(this.next);
		}
		for (Branch branch : this.branches) {
			transitions.add(branch.getInitial());
		}

		return transitions.toArray(new Transition[transitions.size()]);
	}

}
