package display.gui;

import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.DefaultListModel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JTree;
import javax.swing.UIManager;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeCellRenderer;

import counterexample.Branch;
import counterexample.State;
import counterexample.Transition;
import counterexample.Value;

/**
 * A tree info panel is an info panel displaying information about states and
 * transitions in boxes containing trees and lists.
 * 
 * @author Simon Busard <simon.busard@uclouvain.be>
 * 
 */
public class TreeInfoPanel extends InfoPanel {

	private static final long serialVersionUID = 1L;

	/**
	 * Creates a new info panel displaying info into trees in boxes
	 */
	public TreeInfoPanel() {
		super();
	}

	/**
	 * Displays the information contained in the given state.
	 * 
	 * @param state
	 *            the state to display.
	 * @param prevState
	 *            the state before state. If not null and this.changedVar is
	 *            true, then only the state variables that changed during the
	 *            transition between prevState and state are displayed.
	 */
	private void displayStateInfo(State state, State prevState) {
		// Create the panel
		JPanel statePanel = new JPanel();
		statePanel.setBackground(UIManager.getColor("Tree.background"));
		statePanel.setLayout(new BoxLayout(statePanel, BoxLayout.Y_AXIS));
		this.add(statePanel);
		statePanel.setBorder(BorderFactory.createEtchedBorder());
		statePanel.setAlignmentX(1);

		// Create the tree
		DefaultMutableTreeNode rootNode = new DefaultMutableTreeNode();
		// State
		DefaultMutableTreeNode stateNode = new DefaultMutableTreeNode("State "
				+ state.getId());
		rootNode.add(stateNode);
		for (Value value : state.getValues()) {
			if (!this.changedVar
					|| prevState == null
					|| !getValueFromVariable(prevState, value.getVariable())
							.equals(value.getValue())) {
				stateNode.add(new DefaultMutableTreeNode(value));
			}
		}
		// Annotation
		DefaultMutableTreeNode annoNode = new DefaultMutableTreeNode(
				"Annotation");
		rootNode.add(annoNode);
		annoNode.add(new DefaultMutableTreeNode(state.getAnnotation()));

		// Branches
		if (state.getBranches().length > 0) {
			DefaultMutableTreeNode branchesNode = new DefaultMutableTreeNode(
					"Branches");
			rootNode.add(branchesNode);
			for (Branch branch : state.getBranches()) {
				branchesNode.add(new DefaultMutableTreeNode(branch
						.getAnnotation()));
			}
		}

		// Actual tree
		JTree tree = new JTree(rootNode);
		tree.setShowsRootHandles(true);
		tree.setRootVisible(false);
		DefaultTreeCellRenderer cellRenderer = new DefaultTreeCellRenderer();
		cellRenderer.setClosedIcon(null);
		cellRenderer.setLeafIcon(null);
		cellRenderer.setOpenIcon(null);
		tree.setCellRenderer(cellRenderer);

		// Expand state
		tree.expandRow(0);

		// Add the tree to the display
		statePanel.add(tree);
		statePanel.add(Box.createHorizontalGlue());
		tree.setAlignmentX(0);

		this.panel.adjustSplitpane();
	}

	/**
	 * Returns the value of variable of state.
	 * 
	 * @param state
	 *            the state to search.
	 * @param variable
	 *            the variable to search.
	 * @return If state contains a value to the variable variable, returns this
	 *         value, null otherwise.
	 */
	private String getValueFromVariable(State state, String variable) {
		for (Value value : state.getValues()) {
			if (value.getVariable().equals(variable)) {
				return value.getValue();
			}
		}
		return null;
	}

	/**
	 * Displays the information contained in the given transition.
	 * 
	 * @param transition
	 *            the transition to display.
	 */
	private void displayTransitionInfo(Transition transition) {
		if (transition.getInputs().length > 0) {

			JPanel inputPanel = new JPanel();
			inputPanel.setBackground(UIManager.getColor("List.background"));
			this.add(inputPanel);
			inputPanel.setLayout(new BoxLayout(inputPanel, BoxLayout.X_AXIS));
			inputPanel.setBorder(BorderFactory.createEtchedBorder());
			DefaultListModel model = new DefaultListModel();
			for (Value value : transition.getInputs()) {
				model.addElement(value);
			}
			JList input1List = new JList(model);
			inputPanel.add(input1List);
			inputPanel.setAlignmentX(1);

			inputPanel.add(Box.createHorizontalGlue());

			this.panel.adjustSplitpane();
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see display.InfoPanel#displayStatesInfo(java.util.List)
	 */
	@Override
	protected void displayStatesInfo(List<State> states) {
		this.removeAll();
		if (states.size() > 0) {
			displayStateInfo(states.get(0), null);
			State prevState = states.get(0);
			for (int i = 1; i < states.size(); i++) {
				displayTransitionInfo(getTransitionFromOpposite(prevState,
						states.get(i)));
				displayStateInfo(states.get(i), prevState);
				prevState = states.get(i);
			}
		}
	}

	/**
	 * Returns the transition going from source to destination.
	 * 
	 * @param source
	 *            the source of the wished transition.
	 * @param destination
	 *            the destination of the wished transition.
	 * @return the transition between source and destination if any, null
	 *         otherwise.
	 */
	private Transition getTransitionFromOpposite(State source, State destination) {
		for (Transition transition : source.getNexts()) {
			if (transition.getTo().equals(destination)) {
				return transition;
			}
		}
		return null;
	}
}
