package display.layout;

import java.awt.Dimension;
import java.awt.geom.Point2D;
import java.util.HashMap;
import java.util.Map;

import application.Settings;
import counterexample.Counterexample;
import counterexample.State;
import display.graph.Edge;
import display.graph.Node;
import edu.uci.ics.jung.graph.DirectedGraph;
import edu.uci.ics.jung.graph.Graph;

/**
 * The tree like layout layouts a counterexample displaying vertical branches.
 * 
 * @author Simon Busard <simon.busard@uclouvain.be>
 * 
 */
public abstract class TreeLikeLayout extends AbstractCounterexampleLayout
		implements CounterexampleLayout {

	/**
	 * The current displayed graph.
	 */
	protected DirectedGraph<Node, Edge> graph;

	/**
	 * The counterexample represented by graph.
	 */
	protected Counterexample counterexample;

	/**
	 * The current location of nodes;
	 */
	protected Map<Node, Point2D> locations;

	/**
	 * Creates a new layout.
	 */
	public TreeLikeLayout() {
		this.locations = new HashMap<Node, Point2D>();
		this.graph = null;
		this.counterexample = null;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.apache.commons.collections15.Transformer#transform(java.lang.Object)
	 */
	@Override
	public Point2D transform(Node node) {
		return this.locations.get(node);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.uci.ics.jung.algorithms.layout.Layout#getGraph()
	 */
	@Override
	public Graph<Node, Edge> getGraph() {
		return this.graph;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.uci.ics.jung.algorithms.layout.Layout#getSize()
	 */
	@Override
	public Dimension getSize() {
		double minx = Integer.MAX_VALUE;
		double miny = Integer.MAX_VALUE;
		double maxx = Integer.MIN_VALUE;
		double maxy = Integer.MIN_VALUE;

		for (Point2D point : this.locations.values()) {
			if (point.getX() < minx) {
				minx = point.getX();
			} else if (point.getX() > maxx) {
				maxx = point.getX();
			}

			if (point.getY() < miny) {
				miny = point.getY();
			} else if (point.getY() > maxy) {
				maxy = point.getY();
			}
		}

		return new Dimension(2 * Settings.graphPadding
				+ (int) Math.max(0, maxx - minx), 2 * Settings.graphPadding
				+ (int) Math.max(0, maxy - miny));
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * edu.uci.ics.jung.algorithms.layout.Layout#setLocation(java.lang.Object,
	 * java.awt.geom.Point2D)
	 */
	@Override
	public void setLocation(Node node, Point2D location) {
		this.locations.put(node, location);
	}

	/**
	 * Layouts the intern graph by storing each node location in locations.
	 */
	protected abstract void buildLayout();

	/**
	 * Put the given node at location (x,y).
	 * 
	 * @param node
	 *            The given node.
	 * @param x
	 *            The x-axis position.
	 * @param y
	 *            The y-axis position.
	 */
	protected void drawNode(Node node, int x, int y) {
		this.locations.put(node, new Point2D.Double(
				((double) Settings.graphPadding + y * Settings.vertexSpace),
				(double) (Settings.graphPadding + x * Settings.vertexSpace)));
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * display.layout.CounterexampleLayout#initialize(counterexample.Counterexample
	 * , edu.uci.ics.jung.graph.Graph)
	 */
	@Override
	public void initialize(Counterexample counterexample,
			DirectedGraph<Node, Edge> graph) {
		this.counterexample = counterexample;
		this.graph = graph;
		this.locations = new HashMap<Node, Point2D>();

		this.buildLayout();
	}

	/**
	 * Returns the node representing the given state in graph.
	 * 
	 * @param graph
	 *            The graph containing the node.
	 * @param state
	 *            The given state.
	 * @return the node representing state in graph if any, null otherwise.
	 */
	protected Node getNodeFromState(DirectedGraph<Node, Edge> graph, State state) {
		for (Node node : graph.getVertices()) {
			if (node.getState().equals(state)) {
				return node;
			}
		}
		return null;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see display.layout.AbstractCounterexampleLayout#reset()
	 */
	@Override
	public void reset() {
		this.buildLayout();
	}
}
