package display.layout;

import util.ExpandableList;
import util.ExpandableMap;
import counterexample.Branch;
import counterexample.State;
import counterexample.Transition;
import display.graph.Edge;
import display.graph.Node;
import edu.uci.ics.jung.graph.DirectedGraph;

/**
 * The wide tree like layout layouts a counterexample displaying vertical
 * branches.
 * 
 * @author Simon Busard <simon.busard@uclouvain.be>
 * 
 */
public class WideTreeLikeLayout extends TreeLikeLayout implements
		CounterexampleLayout {

	/**
	 * Layouts the intern graph by storing each node location in locations.
	 */
	protected void buildLayout() {
		ExpandableMap<Node> map = new ExpandableMap<Node>();
		ExpandableList<Node> drawn = new ExpandableList<Node>();
		Node node = getNodeFromState(graph, counterexample.getInitial());
		drawNode(node, 0, 0);
		map.set(0, 0, node);
		drawn.set(counterexample.getInitial().getId(), node);
		if (counterexample.getInitial().getNext() != null) {
			drawBranch(graph, counterexample.getInitial().getNext(), 1, 0, map,
					drawn);
			if (counterexample.getInitial().getNext().getTo() != counterexample
					.getInitial()) {
				// Nothing to do
			}
		}
		for (Branch branch : counterexample.getInitial().getBranches()) {
			drawBranch(graph, branch.getInitial(), 1, 1, map, drawn);
			if (branch.getInitial().getTo() != counterexample.getInitial()) {
				// Nothing to do
			}
		}
	}

	/**
	 * Draw a straight vertical branch starting at transition, on graph,
	 * starting at position (x,y).
	 * 
	 * @param graph
	 *            the graph to draw the branch on.
	 * @param transition
	 *            the first transition of the branch.
	 * @param x
	 *            the row of the start position of the branch.
	 * @param y
	 *            the column of the start position of the branch.
	 * @param map
	 *            a map of each drawn vertex (indices are positions on the
	 *            graph).
	 * @param drawn
	 *            a list of each drawn vertex (indices are ids).
	 * @return the column at which the branch has been drawn.
	 */
	private int drawBranch(DirectedGraph<Node, Edge> graph,
			Transition transition, int x, int y, ExpandableMap<Node> map,
			ExpandableList<Node> drawn) {
		while (overlap(transition, map, x, y)) {
			y++;
		}
		Transition current = transition;
		State last = transition.getTo();
		boolean first = true;
		while (current != null) {
			if (current.getTo().getId() <= current.getFrom().getId()) {
				// Loop
				// Nothing to do
				current = null;
			} else {
				// Normal node
				Node node = getNodeFromState(graph, current.getTo());
				if (node != null) {
					drawNode(node, x, y);
					map.set(x, y, node);
					drawn.set(current.getTo().getId(), node);
				}
				if (!first) {
					// Nothing to do
				}

				last = current.getTo();
				current = last.getNext();
				x++;
			}
			first = false;

		}
		x--;
		while (last != transition.getFrom()) {
			if (getNodeFromState(graph, last) != null) {
				for (Branch branch : last.getBranches()) {
					drawBranch(graph, branch.getInitial(), x + 1, y + 1, map,
							drawn);
					if (branch.getInitial().getTo() != last) {
						// Nothing to do
					}
				}
			}
			x--;
			last = last.getPrev().getFrom();
		}

		return y;
	}

	/**
	 * Returns whether the branch starting at transition overlaps another branch
	 * in map if it starts at (x,y).
	 * 
	 * @param transition
	 *            the starting transition of the branch.
	 * @param map
	 *            the reference map.
	 * @param x
	 *            the row of the start position of the branch.
	 * @param y
	 *            the column of the start position of the branch.
	 * @return true if the branch starting at transition overlaps another branch
	 *         in map if it starts at (x,y), false otherwise.
	 */
	private boolean overlap(Transition transition, ExpandableMap<Node> map,
			int x, int y) {
		while (transition != null
				&& transition.getTo().getId() > transition.getFrom().getId()) {
			if (map.get(x, y) != null) {
				return true;
			}
			for (int ny = y; ny < map.height(); ny++) {
				if (map.get(x, ny) != null) {
					return true;
				}
			}
			transition = transition.getTo().getNext();
			x++;
		}
		return false;
	}
}
