package display.listener;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JMenuItem;

import util.ObservableList;
import counterexample.Branch;
import counterexample.State;
import counterexample.Transition;
import display.graph.Edge;
import display.graph.Node;
import edu.uci.ics.jung.graph.Graph;
import edu.uci.ics.jung.visualization.VisualizationViewer;

/**
 * A branch listener listens to the branch part of the popup menu of a node. It
 * shows and hides branches and
 * 
 * @author Simon Busard <simon.busard@uclouvain.be>
 * 
 */
public class BranchListener implements ActionListener {

	/**
	 * The graph this listener acts on.
	 */
	private Graph<Node, Edge> graph;
	/**
	 * The node concerned by this listener.
	 */
	private Node node;
	/**
	 * The list of selected nodes of the graph. Used to highlight nodes.
	 */
	private ObservableList<Node> list;
	/**
	 * The listened menu item responsible for visibility of the branch.
	 */
	private JMenuItem showHide;
	/**
	 * The listened menu item responsible for focusing on the branch.
	 */
	private JMenuItem see;
	/**
	 * The concerned branch.
	 */
	private Branch branch;
	/**
	 * The visualization viewer the graph is painted on.
	 */
	private VisualizationViewer<Node, Edge> vv;

	/**
	 * Creates a new branch listener.
	 * 
	 * @param graph
	 *            the graph to act on.
	 * @param node
	 *            the concerned node.
	 * @param branch
	 *            the concerned branch.
	 * @param list
	 *            the list of selected nodes of the graph. Used to highlight
	 *            nodes.
	 * @param showHide
	 *            the listened menu item responsible for visibility of the
	 *            branch.
	 * @param see
	 *            the listened menu item responsible for focusing on the branch.
	 * @param vv
	 *            the visualization viewer the graph is painted on.
	 */
	public BranchListener(Graph<Node, Edge> graph, Node node, Branch branch,
			ObservableList<Node> list, JMenuItem showHide, JMenuItem see,
			VisualizationViewer<Node, Edge> vv) {
		this.graph = graph;
		this.node = node;
		this.branch = branch;
		this.list = list;
		this.showHide = showHide;
		this.see = see;
		this.vv = vv;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
	 */
	@Override
	public void actionPerformed(ActionEvent e) {
		if (e.getSource().equals(this.showHide)) {
			// Show or hide the branch
			if (this.node.isDisplayed(this.branch)) {
				// The branch is displayed, hide it
				hideBranch(this.graph, this.branch);
				node.setDisplayed(this.branch, false);
				this.showHide.setText("Show");
			} else {
				// The branch is hidden, show it
				showBranch(this.graph, this.node, this.branch);
				this.node.setDisplayed(this.branch, true);
				this.showHide.setText("Hide");
			}

			this.list.notifyObservers();

			this.vv.getGraphLayout().reset();
			this.vv.repaint();
		} else if (e.getSource().equals(this.see)) {
			// Focus on the branch

			// If hidden, show the branch
			if (!this.node.isDisplayed(this.branch)) {
				showBranch(this.graph, this.node, this.branch);
				this.node.setDisplayed(this.branch, true);

				this.vv.getGraphLayout().reset();
				this.vv.repaint();
			}

			// Sets the set of selected nodes to the set of nodes of the branch
			List<Node> nodes = getNodesFromBranch(this.graph, this.branch);
			this.list.clear();
			this.list.add(node);
			for (Node node : nodes) {
				if (node != null) {
					this.list.add(node);
				}
			}
			this.list.notifyObservers();
		}
	}

	/**
	 * Returns the list of nodes representing the states of the given branch in
	 * graph.
	 * 
	 * @param graph
	 *            the graph containing the nodes.
	 * @param branch
	 *            the given branch.
	 * @return the list of nodes representing the states of the given branch in
	 *         graph. If the branch is composed of one single loop, the returned
	 *         list is empty.
	 */
	private List<Node> getNodesFromBranch(Graph<Node, Edge> graph, Branch branch) {
		List<Node> nodes = new ArrayList<Node>();
		for (State state : branch.getStates()) {
			nodes.add(getNodeFromState(graph, state));
		}
		return nodes;
	}

	/**
	 * Returns the list of nodes representing the states of the given branch in
	 * graph. All the nodes of the branch and of sub-branches are returned.
	 * 
	 * @param graph
	 *            the graph containing the nodes.
	 * @param branch
	 *            the given branch.
	 * @return the list of nodes representing the states of the given branch in
	 *         graph. If the branch is composed of one single loop, the returned
	 *         list is empty.
	 */
	private List<Node> getAllNodesFromBranch(Graph<Node, Edge> graph,
			Branch branch) {
		List<Node> nodes = new ArrayList<Node>();
		for (State state : branch.getAllStates()) {
			nodes.add(getNodeFromState(graph, state));
		}
		return nodes;
	}

	/**
	 * Returns the node representing the given state in graph.
	 * 
	 * @param graph
	 *            the graph containing the nodes.
	 * @param state
	 *            the given state.
	 * @return the node representing the given state in graph if any, null
	 *         otherwise.
	 */
	private Node getNodeFromState(Graph<Node, Edge> graph, State state) {
		for (Node node : graph.getVertices()) {
			if (node.getState().equals(state)) {
				return node;
			}
		}
		return null;
	}

	/**
	 * Shows the given branch by adding nodes representing the states of branch,
	 * and linked to node.
	 * 
	 * @param graph
	 *            the graph in which add the nodes.
	 * @param node
	 *            the source node of the branch.
	 * @param branch
	 *            the branch to show.
	 */
	private void showBranch(Graph<Node, Edge> graph, Node node, Branch branch) {
		if (branch.getInitial().getTo().equals(branch.getInitial().getFrom())) {
			// A loop
			graph.addEdge(new Edge(branch.getInitial()),
					getNodeFromState(graph, branch.getInitial().getFrom()),
					getNodeFromState(graph, branch.getInitial().getTo()));
		} else {
			// Add nodes and edges simutaneously
			Node prevNode = node;
			for (State state : branch.getStates()) {
				Node newNode = new Node(state);
				graph.addVertex(newNode);
				graph.addEdge(new Edge(state.getPrev()), prevNode, newNode);
				prevNode = newNode;
			}
		}
	}

	/**
	 * Hides the given branch by removing all the nodes representing states of
	 * branch (and sub-branches) from graph.
	 * 
	 * @param graph
	 *            the graph containing the nodes.
	 * @param branch
	 *            the branch to hide.
	 */
	private void hideBranch(Graph<Node, Edge> graph, Branch branch) {
		if (branch.getInitial().getTo().equals(branch.getInitial().getFrom())) {
			// Single loop : remove the transition
			graph.removeEdge(getEdgeFromTransition(graph, branch.getInitial()));
		} else {
			// Some states are involved : remove the states (the transitions
			// will be automatically removed)
			for (Node node : getAllNodesFromBranch(graph, branch)) {
				graph.removeVertex(node);
				this.list.remove(node);
			}
		}
	}

	/**
	 * Returns the edge of the given graph that represents transition.
	 * 
	 * @param graph
	 *            the given graph.
	 * @param transition
	 *            the given transition.
	 * @return the edge of graph representing transition if any, null otherwise.
	 */
	private Edge getEdgeFromTransition(Graph<Node, Edge> graph,
			Transition transition) {
		for (Edge edge : graph.getEdges()) {
			if (edge.getTransition().equals(transition)) {
				return edge;
			}
		}
		return null;
	}
}
