package display.mouse;

import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.util.HashMap;
import java.util.Map;

import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JSeparator;

import util.ObservableList;

import counterexample.Branch;
import counterexample.Value;
import display.graph.Edge;
import display.graph.Node;
import display.listener.BranchListener;
import edu.uci.ics.jung.algorithms.layout.GraphElementAccessor;
import edu.uci.ics.jung.graph.Graph;
import edu.uci.ics.jung.visualization.VisualizationViewer;
import edu.uci.ics.jung.visualization.control.AbstractPopupGraphMousePlugin;

/**
 * The node popup menu graph mouse plugin displays a popup when clicking with
 * right button on a node. This menu displays the list of state values of this
 * node, its annotation and the list of its branches. Each branch can be focused
 * and shown or hidden depending on its current visibility.
 * 
 * @author Simon Busard <simon.busard@uclouvain.be>
 * 
 */
public class NodePopupMenuGraphMousePlugin extends
		AbstractPopupGraphMousePlugin {

	/**
	 * A map of node popup menus.
	 */
	private Map<Node, JPopupMenu> menus;

	/**
	 * The list of selected nodes, used to initialize branch listeners.
	 */
	private ObservableList<Node> list;

	/**
	 * Creates a new node popup graph mouse plugin based on list.
	 * 
	 * @param list
	 *            a list of selected nodes.
	 */
	public NodePopupMenuGraphMousePlugin(ObservableList<Node> list) {
		this.menus = new HashMap<Node, JPopupMenu>();
		this.list = list;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * edu.uci.ics.jung.visualization.control.AbstractPopupGraphMousePlugin#
	 * handlePopup(java.awt.event.MouseEvent)
	 */
	@SuppressWarnings("unchecked")
	@Override
	protected void handlePopup(MouseEvent event) {
		if (event.getSource() instanceof VisualizationViewer) {
			VisualizationViewer<Node, Edge> vv = (VisualizationViewer<Node, Edge>) event
					.getSource();
			GraphElementAccessor<Node, Edge> pickSupport = vv.getPickSupport();

			Node node = pickSupport.getVertex(vv.getGraphLayout(), event
					.getPoint().getX(), event.getPoint().getY());

			Graph<Node, Edge> graph = vv.getGraphLayout().getGraph();

			// A node has been clicked
			if (node != null) {
				if (!menus.containsKey(node)) {
					// Create the popupMenu
					JPopupMenu popupMenu = new JPopupMenu(Integer.toString(node
							.getState().getId()));

					// Display state values
					for (Value value : node.getState().getValues()) {
						JMenuItem item = new JMenuItem(value.toString());
						item.setEnabled(false);
						popupMenu.add(item);
					}

					// Add annotation
					popupMenu.add(new JSeparator());
					JMenu annotationMenu = new JMenu("Annotation");
					JMenuItem annotationItem = new JMenuItem(node.getState()
							.getAnnotation());
					annotationItem.setEnabled(false);
					annotationMenu.add(annotationItem);
					popupMenu.add(annotationMenu);

					// Add branches
					if (node.getState().getBranches().length > 0) {
						popupMenu.add(new JSeparator());
						JMenu branchMenu = new JMenu("Branches");
						popupMenu.add(branchMenu);
						for (Branch branch : node.getState().getBranches()) {
							JMenu menu = new JMenu(branch.getAnnotation());
							JMenuItem showHide = new JMenuItem(
									(node.isDisplayed(branch) ? "Hide" : "Show"));
							JMenuItem see = new JMenuItem("See");
							menu.add(showHide);
							menu.add(see);

							// Add listeners
							ActionListener listener = new BranchListener(graph,
									node, branch, list, showHide, see, vv);
							showHide.addActionListener(listener);
							see.addActionListener(listener);

							branchMenu.add(menu);
						}
					}

					menus.put(node, popupMenu);
				}

				menus.get(node).show(vv, event.getX(), event.getY());
			}

			event.consume();
		}
	}
}
