package xml;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import util.ExpandableList;
import counterexample.Branch;
import counterexample.Counterexample;
import counterexample.State;
import counterexample.Transition;
import counterexample.Value;

/**
 * XMLLoader gives an internal representation of a trace stored in an XML
 * format.
 * 
 * @author Simon Busard <simon.busard@uclouvain.be>
 * 
 */

public class XMLLoader {

	/**
	 * An internal XPath evaluator used to evaluate XPath expressions on XML
	 * representation of counterexamples.
	 */
	private XPath xpath = XPathFactory.newInstance().newXPath();

	/**
	 * Returns an internal representation of a counterexample stored in an XML
	 * file at filepath.
	 * 
	 * @param filepath
	 *            the path to the file containing the counterexample.
	 * @return an internal representation of the counterexample in the XML file
	 *         at filepath.
	 * @throws IOException
	 *             an IOException is thrown when an error occurred while reading
	 *             the given file.
	 * @throws SAXException
	 *             an SAXException is thrown when an error occurred while
	 *             parsing the given file.
	 */
	public Counterexample getCounterexample(String filepath)
			throws SAXException, IOException {
		Counterexample cntex = this.getCounterexample(new File(filepath));
		return cntex;
	}

	/**
	 * Returns an internal representation of a counterexample stored in a given
	 * XML file.
	 * 
	 * @param file
	 *            the file containing the counterexample.
	 * @return an internal representation of the counterexample stored in file.
	 * @throws IOException
	 *             an IOException is thrown when an error occurred while reading
	 *             the given file.
	 * @throws SAXException
	 *             an SAXException is thrown when an error occurred while
	 *             parsing the given file.
	 */
	public Counterexample getCounterexample(File file) throws SAXException,
			IOException {

		// Open file and get content
		// Gets document builder
		DocumentBuilder builder;
		try {
			builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
			// Reads the file
			Document document = builder.parse(file);

			// Get counterexample
			return this.getCounterexample(document.getDocumentElement());
		} catch (ParserConfigurationException e) {
			System.err
					.println("XMLLoader.getCounterexample : Problem with parser configuration.");
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Returns an internal representation of a counterexample given in an XML
	 * format.
	 * 
	 * @param xml
	 *            the XML format of the counterexample.
	 * @return an internal representation of a counterexample given in XML
	 *         format in xml.
	 */
	public Counterexample getCounterexample(Element xml) {
		try {
			ExpandableList<State> states = new ExpandableList<State>();
			State prev = null, state = null, initial = null;

			// Get nodes
			NodeList nodes = (NodeList) xpath.evaluate("node", xml,
					XPathConstants.NODESET);
			Element node;
			for (int i = 0; i < nodes.getLength(); i++) {
				node = (Element) nodes.item(i);

				// Check the loop
				boolean isLoop = (Boolean) xpath.evaluate("loop/@to", node,
						XPathConstants.BOOLEAN);
				if (isLoop) {
					// Manage the loop
					prev.setNext(new Transition(getInputs(node), prev, states
							.get(Integer.parseInt(xpath.evaluate("loop/@to",
									node)))));
				} else {
					// Get normal state
					state = getState(node, states);
					if (initial == null) {
						initial = state;
					}
					if (prev != null) {
						Transition t = new Transition(getInputs(node), prev,
								state);
						prev.setNext(t);
						state.setPrev(t);
					}

					prev = state;
				}
			}
			return new Counterexample(xpath.evaluate("@specification", xml),
					initial, states.fillArray(new State[states.size()]));
		} catch (XPathExpressionException e) {
			System.err
					.println("XMLLoader.getCounterexample() : Wrong XPath expression.");
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Returns an internal representation of the state stored in node.
	 * 
	 * @param node
	 *            the XML representation of the wanted state.
	 * @param states
	 *            the list of all states already met during the extraction of
	 *            the counterexample that the state belongs to.
	 * @return an internal representation of the state stored in node.
	 */
	private State getState(Element node, ExpandableList<State> states) {
		try {
			// Get variables
			List<Value> variables = new ArrayList<Value>();

			NodeList values = (NodeList) xpath.evaluate("state/value", node,
					XPathConstants.NODESET);
			for (int i = 0; i < values.getLength(); i++) {
				variables.add(new Value(xpath.evaluate("@variable",
						values.item(i)), xpath.evaluate(".", values.item(i))));
			}

			// Create state
			int stateId = Integer.parseInt(xpath.evaluate("state/@id", node));

			State state = new State(stateId,
					variables.toArray(new Value[variables.size()]),
					xpath.evaluate("annotation", node));

			states.set(stateId, state);

			// Get branches
			NodeList branches = (NodeList) xpath.evaluate("branch", node,
					XPathConstants.NODESET);
			for (int i = 0; i < branches.getLength(); i++) {
				state.addBranch(getBranch((Element) branches.item(i), state,
						states));
			}

			return state;
		} catch (XPathExpressionException e) {
			System.err
					.println("XMLLoader.getState() : Wrong XPath expression.");
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Returns an internal representation of the branch stored in branch,
	 * starting at initial. Each new state of the branch is added to states.
	 * 
	 * @param branch
	 *            the XML node representing the branch.
	 * @param initial
	 *            the initial state of the branch.
	 * @param states
	 *            the list of states already created during the extraction of
	 *            the counterexample that the branch belongs to.
	 * @return an internal representation of the branch stored in branch.
	 */
	private Branch getBranch(Element branch, State initial,
			ExpandableList<State> states) {
		try {
			State prev = null;
			Transition initialTrans = null;

			// Get nodes
			NodeList nodes = (NodeList) xpath.evaluate("node", branch,
					XPathConstants.NODESET);
			Element node;
			for (int i = 0; i < nodes.getLength(); i++) {
				node = (Element) nodes.item(i);

				// Check the loop
				boolean isLoop = (Boolean) xpath.evaluate("loop/@to", node,
						XPathConstants.BOOLEAN);
				if (isLoop) {
					// Manage the loop
					if (initialTrans == null) {
						initialTrans = new Transition(getInputs(node), initial,
								states.get(Integer.parseInt(xpath.evaluate(
										"loop/@to", node))));
					} else {
						prev.setNext(new Transition(getInputs(node), prev,
								states.get(Integer.parseInt(xpath.evaluate(
										"loop/@to", node)))));
					}
				} else {
					// Get normal state
					State state = getState(node, states);
					if (initialTrans == null) {
						initialTrans = new Transition(getInputs(node), initial,
								state);
						state.setPrev(initialTrans);
					} else {
						Transition t = new Transition(getInputs(node), prev,
								state);
						prev.setNext(t);
						state.setPrev(t);
					}
					prev = state;
				}
			}
			return new Branch(xpath.evaluate("@specification", branch),
					initialTrans);
		} catch (XPathExpressionException e) {
			System.err
					.println("XMLLoader.getBranch() : Wrong XPath expression.");
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Returns the list of variable-value pairs stored as inputs in node.
	 * 
	 * @param node
	 *            the XML node containing the inputs.
	 * @return an array of inputs (variable-value pairs).
	 */
	private Value[] getInputs(Element node) {
		List<Value> values = new ArrayList<Value>();

		try {
			// Get inputs
			NodeList inputs = (NodeList) xpath.evaluate("input/value", node,
					XPathConstants.NODESET);

			for (int j = 0; j < inputs.getLength(); j++) {
				values.add(new Value(
						xpath.evaluate("@variable", inputs.item(j)), xpath
								.evaluate(".", inputs.item(j))));
			}

			// Get combinatorial
			NodeList combinatorial = (NodeList) xpath.evaluate(
					"combinatorial/value", node, XPathConstants.NODESET);

			for (int j = 0; j < combinatorial.getLength(); j++) {
				values.add(new Value(xpath.evaluate("@variable",
						combinatorial.item(j)), xpath.evaluate(".",
						combinatorial.item(j))));
			}
		} catch (XPathExpressionException e) {
			System.err
					.println("XMLLoader.getInputs() : Wrong XPath expression.");
			e.printStackTrace();
		}

		return values.toArray(new Value[values.size()]);
	}
}
